# inspired from:
# https://python-packaging-user-guide.readthedocs.org/en/latest/single_source_version/
# https://github.com/pypa/warehouse/blob/master/warehouse/__about__.py
# https://github.com/explosion/spaCy/blob/master/spacy/about.py

__title__ = 'medspacy_quickumls'
__version__ = '2.2'
__author__ = 'Luca Soldaini, modified by Kelly Peterson'
__email__ = 'luca@ir.cs.georgetown.edu,kelly.peterson@hsc.utah.edu'
__license__ = 'MIT'
__uri__ = "https://github.com/medspacy/quickumls"
__copyright__ = '2014-2020, Georgetown University Information Retrieval Lab, updated 2020, University of Utah'
